/* Generated by Cython 3.0.11 */

#ifndef __PYX_HAVE_API__CyRK__cy__pysolver_cyhook
#define __PYX_HAVE_API__CyRK__cy__pysolver_cyhook
#ifdef __MINGW64__
#define MS_WIN64
#endif
#include "Python.h"
#include "pysolver_cyhook.h"

static void (*__pyx_api_f_4CyRK_2cy_15pysolver_cyhook_call_diffeq_from_cython)(PyObject *, DiffeqMethod) = 0;
#define call_diffeq_from_cython __pyx_api_f_4CyRK_2cy_15pysolver_cyhook_call_diffeq_from_cython
#ifndef __PYX_HAVE_RT_ImportFunction_3_0_11
#define __PYX_HAVE_RT_ImportFunction_3_0_11
static int __Pyx_ImportFunction_3_0_11(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif


static int import_CyRK__cy__pysolver_cyhook(void) {
  PyObject *module = 0;
  module = PyImport_ImportModule("CyRK.cy.pysolver_cyhook");
  if (!module) goto bad;
  if (__Pyx_ImportFunction_3_0_11(module, "call_diffeq_from_cython", (void (**)(void))&__pyx_api_f_4CyRK_2cy_15pysolver_cyhook_call_diffeq_from_cython, "void (PyObject *, DiffeqMethod)") < 0) goto bad;
  Py_DECREF(module); module = 0;
  return 0;
  bad:
  Py_XDECREF(module);
  return -1;
}

#endif /* !__PYX_HAVE_API__CyRK__cy__pysolver_cyhook */
